structure Norm = NormFunc(structure Term = Term)
structure Unif = HO_Pattern_Unify(structure Term = Norm.Term  
                                  structure Norm = Norm)

structure T = Term

(* EXAMPLE 1
   This one tests a simple case involving abstractions
   unify  lam x x  with  lam x (M x)
   result should be M -> lam x x
*)

val dB1 = ref (T.dB 1)

val t1 = ref (T.lam (1, dB1))

val M = (ref (T.var ("M",1)))

val t2 = ref (T.lam(1,
                    ref (T.app(M, 
                              ((ref (T.dB 1)) :: nil)))))

Unif.pattern_unify t1 t2

M


(* EXAMPLE 2
   Here we add some processing that needs to descend into constructors
   unify  lam x (c x) with lam x (c (N x))
   result should be N -> lam x x
*)

val N = ref (T.var ("N",1))

val t7 = ref (T.lam (1, (ref (T.app ( ref (T.const ("c",1)),
                                      (ref (T.dB 1))::nil)))))

val t8 = ref (T.lam (1, (ref (T.app (ref (T.const ("c",1)),
                                     (ref (T.app (N,
                                                  (ref (T.dB 1))::nil)))::nil)
                             ))))

Unif.pattern_unify t7 t8

N

(* EXAMPLE 3
   Here we also need to eta expand on-the-fly
   unify lam x lam y (c(1) y x) with N(1)
   result should be N -> lam x lam y (c y x)
*)

val N2 = ref (T.var ("N",1))

val t9 = ref (T.lam (2, (ref (T.app ( ref (T.const ("c",1)),
                                      (ref (T.dB 1))::(ref (T.dB 2)):: nil)))))

Unif.pattern_unify N2 t9

(* EXAMPLE 4
   Here we also do eta on-the-fly, but with constructors at the top level
   unify lam x lam y (c(1) x y) with lam x (c(1) (N x))
   result should be N -> lam x x
*)

val N3 = ref (T.var ("N",1))

val t11 = ref (T.lam (2, (ref (T.app ( ref (T.const ("c",1)),
                                      (ref (T.dB 2))::(ref (T.dB 1)):: nil)))))

val t12 = ref (T.lam (1, (ref (T.app (ref (T.const ("c",1)),
                                      (ref (T.app (N3,
                                                   (ref (T.dB 1)):: nil)))
                                          :: nil)))))

Unif.pattern_unify t11 t12

N3



(* EXAMPLE 5
  Here we have a flex-flex case where we need to raise and prune
  unify X(1) a(2) b(3) with Y(2) b(3) c(3)
  result should be X -> lam x y (Z(1) x y) and Y -> lam x y (Z a x)
*)

val X1 = (ref (T.var ("X",1)))

val t3 = ref (T.app (X1,
                    ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val Y1 = (ref (T.var ("Y",2)))

val t4 = ref (T.app (Y1,
                     ((ref (T.const ("b",3)))::(ref (T.const ("c",3)))::nil)))

Unif.pattern_unify t3 t4

X1

Y1


(* EXAMPLE 6
  Here we have a flex-rigid case where we need to raise and prune
  relative to an embedded flex term
; unify X(1) a(2) b(3) with (c(1) (Y(2) b(3) c(3)))
; result should be X -> lam x y (c (Z(1) x y)) and Y -> lam x y (Z a x)
*)

val X2 = (ref (T.var ("X",1)))

val t13 = ref (T.app (X2,
                    ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val Y2 = (ref (T.var ("Y",2)))

val t14 = ref (T.app ((ref (T.const ("c",1))),
                      (ref (T.app (Y2,
                             ((ref (T.const ("b",3)))::
                              (ref (T.const ("c",3)))::nil))))::nil))

Unif.pattern_unify t13 t14

X2

Y2


(* EXAMPLE 7
   Multiple occurrences of the same variable
   unify (c(1) (X(1) a(2) b(3)) (X b(3) d(2))) with 
         (c(1) (Y(2) b(3) c(3)) (b(3) d(2)))
   result should be X -> lam x y (x y) and Y -> lam x y (a x)
*)

val X3 = (ref (T.var ("X",1)))

val Y3 = (ref (T.var ("Y",2)))

val t15 = ref (T.app (ref (T.const ("c",1)),
                      (ref (T.app (X3,
                                   (ref (T.const ("a",2)))::
                                          (ref (T.const ("b",3)))::nil)))::
                      (ref (T.app (X3,
                                   (ref (T.const ("b",3)))::
                                          (ref (T.const ("d",3)))::nil)))::nil)
              )

val t16 = ref (T.app (ref (T.const ("c",1)),
                      (ref (T.app (Y3,
                                   (ref (T.const ("b",3)))::
                                          (ref (T.const ("c",3)))::nil)))::
                      (ref (T.app (ref (T.const ("b",3)),
                                   (ref (T.const ("d",3)))::nil)))::nil))

Unif.pattern_unify t15 t16

X3

Y3



(* EXAMPLE 8
   Multiple occurrences of the same variable with a bound variable
   as the rigid part rather than a constant
   unify lam x (c(1) (X(1) a(2) b(3)) (X x d(2))) with 
         lam x (c(1) (Y(2) b(3) c(3)) (x d(2)))
   result should be X -> lam x y (x y) and Y -> lam x y (a x)
*)

val X4 = (ref (T.var ("X",1)))

val Y4 = (ref (T.var ("Y",2)))

val t17 = ref (T.lam 
             (1,
             (ref (T.app (ref (T.const ("c",1)),
                      (ref (T.app (X4,
                                   (ref (T.const ("a",2)))::
                                          (ref (T.const ("b",3)))::nil)))::
                      (ref (T.app (X4,
                                   (ref (T.dB 1))::
                                          (ref (T.const ("d",3)))::nil)))::nil)
              ))))

val t18 = ref (T.lam
            (1,  
             (ref (T.app (ref (T.const ("c",1)),
                      (ref (T.app (Y4,
                                   (ref (T.const ("b",3)))::
                                          (ref (T.const ("c",3)))::nil)))::
                      (ref (T.app (ref (T.dB 1),
                                   (ref (T.const ("d",3)))::nil)))::nil)))))

Unif.pattern_unify t17 t18

X4

Y4


(* Example 9
   Flex-Flex with the same variable as the head in both terms
   unify X(1) a(2) b(3) c(3)  =  X(1) c(3) b(3) a(2)
   result should be X -> lam x lam y lam z (W(1) y)
*)

val a = ref (T.const("a",2))

val b = ref (T.const("b",3))

val c = ref (T.const("c",3))

val X5 = ref (T.var("X", 1))
val args1 = [a,b,c]
val args2 = [c,b,a]
val t19 = ref (T.app(X5,args1))
val t20 = ref (T.app(X5,args2))

Unif.pattern_unify t19 t20


(* EXAMPLE 10
  Failure due to occurs check
  unify X(1) a(2) b(3) with (c(1) (X(1) b(3) c(3)))
*)

val X6 = (ref (T.var ("X",1)))

val t21 = ref (T.app (X6,
                    ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val t22 = ref (T.app ((ref (T.const ("c",1))),
                      (ref (T.app (X6,
                             ((ref (T.const ("b",3)))::
                              (ref (T.const ("c",3)))::nil))))::nil))

Unif.pattern_unify t21 t22


(* EXAMPLE 10
  Failure because of quantifier dependency violation (also an occurs check 
  failure)
  unify X(1) a(2) b(3) with (c(3) (X(1) b(3) c(3)))
*)

val X6 = (ref (T.var ("X",1)))

val t23 = ref (T.app (X6,
                    ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val t24 = ref (T.app ((ref (T.const ("c",3))),
                      (ref (T.app (X6,
                             ((ref (T.const ("b",3)))::
                              (ref (T.const ("c",3)))::nil))))::nil))

Unif.pattern_unify t23 t24


(* EXAMPLE 11
  Flex-flex case where there is no raising
; unify X(1) a(2) b(3) with Y(1) b(3) c(3)
; result should be X -> lam x y (Z(1) y) and Y -> lam x y (Z(1) x)
*)

val X2 = (ref (T.var ("X",1)))

val t5 = ref (T.app (X2,
                    ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val Y2 = (ref (T.var ("Y",1)))

val t6 = ref (T.app (Y2,
                     ((ref (T.const ("b",3)))::(ref (T.const ("c",3)))::nil)))

Unif.pattern_unify t5 t6

X2

Y2



(* EXAMPLE 12
  Another flex-flex case where there is raising with respect to one 
  variable and pruning with respect to the other
  X1 a2 b3 c3 = Y2 c3
  result should be X -> lam x y z (W x z) and  Y -> lam x (W a x)
*) 

val X7 = ref (T.var("X",1))
val Y7 = ref (T.var("Y",2))
val a = ref (T.const("a",2))
val b = ref (T.const("b",3))
val c = ref (T.const("c",3))
val args1 = [a,b,c]
val args2 = [c]
val t25 = ref (T.app(X7, args1))
val t26 = ref (T.app(Y7, args2))

Unif.pattern_unify t25 t26

X7

Y7 


(* Example 13
  Flex rigid where the rigid head has to be abstracted
  unify  X(1) a(2) b(3)   with   (a(2) (Y(2) b(3) c(3)))
  has solutions  X -> lam x y (x (H(1) x y)),   Y -> lam x y (H(1) a(2) x)
*)

val X8 = ref (T.var("X",1))

val Y8 = ref (T.var("Y",2))

val t27 = ref (T.app (X8,
                      ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val t28 = ref (T.app ((ref (T.const ("a",2))),
                      ((ref (T.app (Y8,
                       ((ref (T.const ("b",3)))::
                            (ref (T.const ("c",3)))::nil))))::nil)))

Unif.pattern_unify t27 t28

X8

Y8

(* Example 14
   Another occurs check failure case 
   unify  X(1) a(2) b(3)   with   (d(3) (Y(2) b(3) c(3)))
*)

val t29 = ref (T.app ((ref (T.var ("X",1))),
                     ((ref (T.const ("a",2)))::(ref (T.const ("b",3)))::nil)))

val t30 = ref (T.app ((ref (T.const ("d",3))),
                      ((ref (T.app ((ref (T.var ("Y",2))),
                                    ((ref (T.const ("b",3)))::
                                       (ref (T.const ("c",3)))::nil))))
                                         ::nil)))

Unif.pattern_unify t29 t30


(* Example 15
   Unifying two constants
   unify  a(1)   with   a(1)
*)

val t31 = ref (T.const ("a",1))

val t32 = ref (T.const ("a",1))

Unif.pattern_unify t31 t32


(* Example 16
   Unifying two rigid terms
   unify  lam x (a(1) x b(1))  with  lam x (a(1) x b(1))
*)

val t33 = ref (T.lam (1,
                      (ref (T.app (ref (T.const ("a",1)),
                                  (ref (T.dB 1))::
                                      (ref (T.const ("b",1)))::nil)))))

val t34 = ref (T.lam (1,
                      (ref (T.app (ref (T.const ("a",1)),
                                  (ref (T.dB 1))::
                                      (ref (T.const ("b",1)))::nil)))))

Unif.pattern_unify t33 t34


(* Example 17
   Checking that index calculation under embedded abstractions is 
   done right; this is a test based on a bug report (and fix) 
   provided by David Baelde)

   unify  lam x lam y (P x) with lam x (Q x)
   should succeed by binding Q to lam w lam v (P w)

val t35 = ref (T.lam (2, 
                      ref (T.app (ref (T.var ("P",1)), [ ref (T.dB 2) ])))) 

val q = ref (T.var ("Q",1)) 

val t36 = ref (T.lam (1, ref (T.app (q, [ ref (T.dB 1) ]))))

Unif.pattern_unify t35 t36

